/*
 *  GGShip.m
 *
 *  Copyright (c) 2001 by Frdric De Jaeger
 *  
 *  Date: Mars 2001
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
  Various notifications
 
- (void) updateShipData :  called sometimes for fuel and so on
- (void) updatePosition : called every frame to update graphical states
- tryGlobalCoordinates \   called sometimes to manage the reference frame
- tryLocalCoordinates  /   of the ship
*/

#include <Foundation/NSObject.h>
#include <Foundation/NSString.h>
#include <Foundation/NSNotification.h>
#include <Foundation/NSException.h>
#include <Foundation/NSDebug.h>
#include <Foundation/NSDictionary.h>
#include <Foundation/NSArray.h>
#include <GL/glu.h>

#include "GGLoop.h"
#include "GGSpaceObject.h"
#include "GGInput.h"
#include "GGRepere.h"
#include "Metaobj.h"
#include "utile.h"
#include "Tableau.h"
#include "Sched.h"
#include "World.h"
#include "Laser.h"
#include "parseASC.h"
#include "Item.h"
#include "GGConcreteShip.h"
#include "Missile.h"
#include "Resource.h"
#include "Console.h"
#include "Preference.h"

#include <math.h>
#include <time.h>

static const float PowerOverConso = 300*20e3*100;
static const float PowerOverType = 300*20e3;

NSString *GGNotificationShipLeaveCurrentSystem = 
@"GGNotificationShipLeaveCurrentSystem";
NSString *GGNotificationShipEnterCurrentSystem =
@"GGNotificationShipEnterCurrentSystem";
NSString *GGNotificationShipBecomeActivated = 
@"GGNotificationShipBecomeActivated";
NSString *GGNotificationShipBecomeInactivated =
@"GGNotificationShipBecomeInactivated";


@interface GGShip (Private)
- (void) _activate;
- (void) setModel: (ModelObject *) _model;
@end


@implementation GGShip
- (void) _GGShipDefaultInitVar
{
  masseCoque = 5e3;
  capaciteFret = 25e3;
  hyperDriveType = 1;
  pylonRocket = 2;
  pylonLaser = 1;
  carburant = 1000;
  fric = 1000;
  currentMissile = -1;
  gAcc = 10;
  gFrein = 3;
  initVect(&posReacteur1, -3.32, 0.44, -2.9);
  initVect(&posReacteur2, 3.32, 0.44, -2.9);
  initVect(&inertie, 1, 1, 1);

  //FIXME maybe it should not be here
  _flags.graphical = YES;
  [self setMode: GGInertialRotationalMode | GGInertialMode];
}

- (void) _updateInternalData
{
  masse = masseFret + masseCoque;
  //  gAcc = gFrein = PowerOverType * hyperDriveType/masse;
  consommation = PowerOverType * hyperDriveType / PowerOverConso;
}

- (void) _computeInternalData
{
  masseFret = 10e3*hyperDriveType;
  /* Compute things according to some goodies...
   */

  /* compute size of object */
  if( model != nil )
    {
      Vect3D min, max, *pmin;
      real r;
      NSDebugMLLog(@"GGEnemi", @"chargement ok de %@");
      
      pmin = [model boundingBox: &max];
      min = *pmin;
      r = fabs(min.x);
      r = (r > fabs(min.y) ? r : fabs(min.y));
      r = (r > fabs(min.z) ? r : fabs(min.z));
      r = (r > fabs(max.x) ? r : fabs(max.x));
      r = (r > fabs(max.y) ? r : fabs(max.y));
      r = (r > fabs(max.z) ? r : fabs(max.z));
      NSDebugMLLog(@"GGEnemi", @"radius = %g", r);
      dimension = r;
    }

  endurance = 1e6;
  bouclier = 1e6;
  timeLaserRelaxation = 1.0;
  [self _updateInternalData];
}


- init
{
  [super init];
  [self _GGShipDefaultInitVar];
  [self _computeInternalData];

  equipement = [NSMutableArray new];
  xMouvement = yMouvement = rolMouvement = 0.0;
  power = 0.0;
  angleManche = 0;
  speedRotation = 1.0;
  powerRotation = 1.0;
  invincible = NO;
  wishDir = *Direction(self);
  _cosam = 1;
  _sinam = 0;
  lastTime = -1.0;
  timeStartFire = -1.0;
  distanceLocal2 = 4*distanceCameraRepere2;
  distanceGlobal2 = 4*distanceLocal2;

  modeInertie = 0;


  return self;
}

- (void) setVarWithDictionary: (NSDictionary *)dico
{
  NSString *val;

  if( (val = [dico objectForKey: @"MasseCoque"] ) != nil )
    {
      masseCoque = [val doubleValue];
    }

  if( (val = [dico objectForKey: @"Capacite"] ) != nil )
    {
      capaciteFret = [val doubleValue];
    }

  if( (val = [dico objectForKey: @"HyperDriveType"] ) != nil )
    {
      hyperDriveType = [val intValue];
    }

  if( (val = [dico objectForKey: @"PylonRocket"] ) != nil )
    {
      pylonRocket = [val intValue];
    }

  if( (val = [dico objectForKey: @"Gun"] ) != nil )
    {
      pylonLaser = [val intValue];
    }

  if( (val = [dico objectForKey: @"MainAcc"] ) != nil )
    {
      gAcc = [val doubleValue]*10;  /*measured in g */
    }

  if( (val = [dico objectForKey: @"RetroAcc"] ) != nil )
    {
      gFrein = [val doubleValue]*10;
    }


  if( (val = [dico objectForKey: @"Modele"] ) != nil )
    {
      ModelObject *m = [ModelObject modelWithFileName: val
				    withDictionary: dico];
      [self setModel: m];
    }

  [self _computeInternalData];
}

- initWithDictionary: (NSDictionary *)dico
{

  [self init];
  [self setVarWithDictionary: dico];
  ASSIGN(property, dico);

  return self;
}

- (void)encodeWithCoder:(NSCoder *)encoder
{

  [super encodeWithCoder: encoder];
  GGEncode(equipement);
  GGEncode(property);
  GGEncode(target);
  GGEncode(ia);
  GGEncode(gAcc);
  GGEncode(gFrein);
  GGEncode(resultante);
  GGEncode(momentum);
  GGEncode(inertie);
  GGEncode(modeInertie);
  GGEncode(angleManche);
  GGEncode(_cosam);
  GGEncode(_sinam);
  GGEncode(speedRotation);
  GGEncode(powerRotation);
  GGEncode(power);
  GGEncode(xMouvement);
  GGEncode(yMouvement);
  GGEncode(masseCoque);
  GGEncode(capaciteFret);
  GGEncode(masseFret);

  GGEncode(hyperDriveType);
  GGEncode(pylonRocket);
  GGEncode(pylonLaser);

  GGEncode(bouclier);
  GGEncode(endurance);
  GGEncode(carburant);
  GGEncode(fric);
  GGEncode(consommation);

  [encoder encodeArrayOfObjCType: @encode(id)
	   count: MAXMISSILE
	   at: &missiles[0]];

  GGEncode(currentLaser);
  GGEncode(currentMissile);
  GGEncode(nbrMissile);
  GGEncode(powerLaser);
  [encoder encodeArrayOfObjCType: @encode(id)
	   count: MAXLASER
	   at: &laserGun[0]];

  GGEncode(timeStartFire);
  GGEncode(timeLaserRelaxation);
  GGEncode(laser);

  GGEncode(lastTime);
  GGEncode(wishDir);
  GGEncode(pousse);

  GGEncode(posReacteur1);
  GGEncode(posReacteur2);
  GGEncode(invincible);
  GGEncode(tracked);
  GGEncode(indexSystem);
}

- (id)initWithCoder:(NSCoder *)decoder
{
  GGShip *_tar;
  
  [super initWithCoder: decoder];
  [self _GGShipDefaultInitVar];

  GGDecode(equipement);
  GGDecode(property);

  [self setVarWithDictionary: property];  //will load the model

  _tar = [decoder decodeObject];  //we don't retain.
  [self setTarget: _tar];

  GGDecode(ia);
  GGDecode(gAcc);
  GGDecode(gFrein);
  GGDecode(resultante);
  GGDecode(momentum);
  GGDecode(inertie);
  GGDecode(modeInertie);
  GGDecode(angleManche);
  GGDecode(_cosam);
  GGDecode(_sinam);
  GGDecode(speedRotation);
  GGDecode(powerRotation);
  GGDecode(power);
  GGDecode(xMouvement);
  GGDecode(yMouvement);
  GGDecode(masseCoque);
  GGDecode(capaciteFret);
  GGDecode(masseFret);

  GGDecode(hyperDriveType);
  GGDecode(pylonRocket);
  GGDecode(pylonLaser);

  GGDecode(bouclier);
  GGDecode(endurance);
  GGDecode(carburant);
  GGDecode(fric);
  GGDecode(consommation);

  [decoder decodeArrayOfObjCType: @encode(id)
	   count: MAXMISSILE
	   at: &missiles[0]];

  GGDecode(currentLaser);
  GGDecode(currentMissile);
  GGDecode(nbrMissile);
  GGDecode(powerLaser);
  [decoder decodeArrayOfObjCType: @encode(id)
	   count: MAXLASER
	   at: &laserGun[0]];

  GGDecode(timeStartFire);
  GGDecode(timeLaserRelaxation);
  GGDecode(laser);

  GGDecode(lastTime);
  GGDecode(wishDir);
  GGDecode(pousse);

  GGDecode(posReacteur1);
  GGDecode(posReacteur2);
  GGDecode(invincible);
  GGDecode(tracked);
  GGDecode(indexSystem);

  return self;
}


- (id) copyWithZone: (NSZone *)zone
{
  NSAssert(0, @"copy with a ship is forbidden");
  abort();
  return nil;
}

- (BOOL) shouldSave
{
  return YES;
}

- (BOOL) saveChannel
{
  return YES;
}


- (ShipErrorCode) buyItem: (Item *)it
{
  ShipErrorCode code;
  if( it->prixAchat <= fric )
    {
      if( (code = [self addItem: it]) == NoError )
	{
	  fric -= it->prixAchat;
	  return NoError;
	}
      else
	return code;
    }
  else
    {
      return NoMoney;
    }
}
- (ShipErrorCode) sellItem: (Item *)it
{
  ShipErrorCode code;
  code = [self removeItem: it];

  if( code != NoError )
    return code;
  else
    {
      fric += it->prixVente;
      return NoError;
    }
}

- (ShipErrorCode) addItem: (Item *)it
{
  int i;
  NSDebugMLLog(@"addItem", @"obj %@ %d", it, it->type);
  if( it->masse + masseFret > capaciteFret )
    return NoPlaceLeft;
  switch(it->type)
    {
    case LaserPulsion:
      NSAssert(it->arg2 >= 0 && it->arg2 < pylonLaser,
	       NSInternalInconsistencyException);
      if( laserGun[it->arg2] == nil )
	{
	  ASSIGN(laserGun[it->arg2], it);
	  //FIXME: not necessary
	  powerLaser = it->arg1*1e6;
	  timeLaserRelaxation = 1.0;
	}
      else
	{
	  return GunBusy;
	}
      break;
    case LaserRayon:
      NSAssert(it->arg2 >= 0 && it->arg2 < pylonLaser,
	       NSInternalInconsistencyException);
      if( laserGun[it->arg2] == nil )
	{
	  ASSIGN(laserGun[it->arg2], it);
	  //FIXME: not necessary
	  powerLaser = it->arg1*1e6;
	  timeLaserRelaxation = 0.0;
	}
      else
	{
	  return GunBusy;
	}
      break;
    case Missile:
      for( i = 0; i < pylonRocket; ++i )
	{
	  if( missiles[i] == nil )
	    {
	      ASSIGN(missiles[i], it);
	      nbrMissile++;
	      break;
	    }
	}
      if( i == pylonRocket )
	return MissilePylonBusy;
      break;
    case Mine:
      break;
    case BombeEnergie:
      break;
    case ECM:
      break;
    default:
      break;
    }
  [equipement addObject: it];
  masseFret += it->masse;
  [self _updateInternalData];
  return NoError;
}

- (ShipErrorCode) removeItem: (Item *)it
{
  int index;
  int i;
  index = [equipement indexOfObject: it];
  if( index == NSNotFound )
    return EquipmentNotFound;
  
  [equipement removeObjectAtIndex: index];
  masseFret -=  it->masse;
  [self _updateInternalData];
  
  switch(it->type)
    {
    case LaserRayon:
    case LaserPulsion:
      for( i = 0; i < MAXLASER; ++i)
	{
	  if( [laserGun[i] isEqual: it] )
	    {
	      DESTROY(laserGun[i]);
	      break;
	    }
	}
      break;
    case Missile:
      for( i = 0; i < MAXMISSILE; ++i)
	{
	  if( [missiles[i] isEqual: it] )
	    {
	      DESTROY(missiles[i]);
	      nbrMissile--;
	      break;
	    }
	}
      break;
    default:
      break;
    }
  return NoError;
}

- (int) nbrItem: (Item *)it
{
  int n = [equipement count];
  int ret = 0;
  if( n > 0 )
    {
      Item *items[n];
      int i;
      [equipement getObjects: items];

      for( i = 0; i < n; ++i )
	{
	  if( [it isEqual: items[i]] )
	    ret++;
	}
    }
  return ret;
}

- (float) rangeForHyperJump
{
  int nH = [self nbrItem: [Item itemWithType: Hydrogen]];
  return nH*5;
}

static NSString *randomMatricule(void)
{
  return [NSString stringWithFormat: 
		     @"%c%c-%d %c",'A'+irnd(26), 'A'+irnd(26), irnd(10000),
		   'A'+irnd(26)];
}

+ shipWithFileName: (NSString *)name
{
  NSDictionary *dic;
  name = [[theResourceManager modelPath] stringByAppendingPathComponent:
					   name];
  dic = [NSDictionary dictionaryWithContentsOfFile: name];
  if( dic == nil )
    {
      NSWarnMLog(@"can't init ship with %@", name);
      return nil;
    }
  else
    {
      GGShip *aShip = [self alloc];
      [aShip initWithDictionary: dic];
      [aShip setName: randomMatricule()];
      return AUTORELEASE(aShip);
    }
}

- (NSObject<IA> *) ia
{
  return ia;
}

- (void) setIA: (NSObject <IA> *)aIa
{
  ASSIGN(ia, aIa);
  if( ia != nil )
    {
      [MessageManager 
	addObserver: self
	selector: @selector(atDestination:)
	name: GGNotificationArriveToStation
	object: self];
    }
}

- (GGShip *) target
{
  return target;
}

- (void) setTarget: (GGSpaceObject *) newTarget
{
  if( newTarget == target )
    return;
  if( tracked && target != nil && [target isKindOfClass: [GGShip class]] )
    {
      [(GGShip *)target setTrack: NO];
    }

  target =  (GGShip *)newTarget;
  [newTarget enregistreForRemoving: self
	     with: @selector(removeTargetWithNotification:)];

  if( tracked && target != nil && [target isKindOfClass: [GGShip class]] )
    {
      [(GGShip *)target setTrack: YES];
    }
}

- (void) notify: (NSString *) format,...
{
}

- (void) removeTargetWithNotification:  (NSNotification *) aNotification
{
  if( [aNotification object] == target )
    {
      [self setTarget: nil];
      [self notify: @"Target lost!!"];
    }
}

- (void) dealloc
{
  int i;
  for( i = 0; i < MAXMISSILE; ++i )
    RELEASE(missiles[i]);
  for( i = 0; i <MAXLASER ; ++i)
    RELEASE(laserGun[i]);
  RELEASE(model);
  RELEASE(ia);
  RELEASE(equipement);
  RELEASE(laser);
  RELEASE(property);
  [super dealloc];
}

- (void) destroy
{
  //  [ia setShip: nil];
  [super destroy];
}

- (float) power
{
  return power;
}

- (void) setPower: (float) newPower
{
  _flags.stateChange = YES;
  if( carburant <= 0 )
    return;
  if( newPower > 1 )
    power = 1.0;
  else if (newPower < 0)
    power = 0.0;
  else
    power = newPower;
}

- (void) atDestination: (NSNotification *)aNotification
{
  [self stopAutoPilot];
}

- (void) looseTarget: (GGSpaceObject *)obj
{
  [self stopAutoPilot];
  [self notify: @"target %@ lost!", obj];
  [self setTarget: nil];
}


- (void) stopEngine
{
  [self stopAutoPilot];
  [self stop];
}

- (void) stop
{
  modeInertie = 0;
  [self setPower: 0.0];
}

- (void) stopAutoPilot
{
  DESTROY(ia);
  [self stop];
}

- (void) setXMouvement: (float) x
{
  xMouvement = x;
}

- (void) setYMouvement: (float) y
{
  yMouvement = y;
}

- (void) setRolMouvement: (float) val
{
  rolMouvement = val;
}

- (void) turnLeft
{
  modeInertie |= GGLeft;
}

- (void) stopTurnLeft
{
  modeInertie &= ~GGLeft;
}

- (void) turnRight
{
  modeInertie |= GGRight;
}

- (void) stopTurnRight
{
  modeInertie &= ~GGRight;
}

- (void) turnUp
{
  modeInertie |= GGUp;
}

- (void) stopTurnUp
{
  modeInertie &= ~GGUp;
}

- (void) turnDown
{
  modeInertie |= GGDown;
}

- (void) stopTurnDown
{
  modeInertie &= ~GGDown;
}

- (void) rolLeft
{
  modeInertie |= GGStrafLeft;
}
- (void) stopRolLeft
{
  modeInertie &= ~GGStrafLeft;
}

- (void) rolRight
{
  modeInertie |= GGStrafRight;
}
- (void) stopRolRight
{
  modeInertie &= ~GGStrafRight;
}

- (void) accelere
{
  modeInertie |= GGAccelere;
  [self setPower: 1.0];
}

- (void) stopAccelere
{
  modeInertie &= ~GGAccelere;
  [self setPower: 0.0];
}

- (void) decelere
{
  modeInertie |= GGDecelere;
  [self setPower: 1.0];
}

- (void) stopDecelere
{
  modeInertie &= ~GGDecelere;
  [self setPower: 0.0];
}


- (void) startManche
{
  modeInertie |= GGManche;
}

- (void) stopManche
{
  modeInertie &= ~GGManche;
}


- (void) setPousse: (Vect3D *)dir
{
  pousse = *dir;
  modeInertie |= GGPousse;
}

- (void) stopPousse
{
  modeInertie &= ~GGPousse;
  [self setPower: 0.0];
}

- (void) setAngleManche: (float) alpha
{
  angleManche = alpha;
  _cosam = cos(alpha);
  _sinam = sin(alpha);
}

- (void) startWishDirection
{
  modeInertie |= GGWishDirection;
}

- (void) stopWishDirection
{
  modeInertie &= ~GGWishDirection;
}

- (void) setWishDirection: (Vect3D *)vect
	      speedManche:(float) val
  /* this imply starWishDirection*/
{
  wishDir = *vect;
  normaliseVect(&wishDir);
  modeInertie |= GGWishDirection;
  [self setSpeedManche: val];
}

- (void) setSpeedManche: (float) val
{
  if( val < 0 )
    val = 0;
  else if( val > 1 )
    val = 1;
  powerRotation = val;
}

- (void) _adjustMatrixWithDirection: (Vect3D *) pdir
			   orthoDir: (Vect3D *) pv
			       cosb: (float) cosb
			       sinb: (float) sinb
{
  Vect3D 		axe, haut;
  real			x, y;


  prodVect(Direction(self), pv, &axe);

  /*Direction(self), v, axe forment un repre orthonorm*/

  x = prodScal(Haut(self), &axe);
  y = prodScal(Haut(self), pv);
  mulScalVect(&axe, x, &haut);
  addLambdaVect(&haut, cosb*y, pv, &haut);
  addLambdaVect(&haut, -sinb*y, Direction(self), &haut);
	  
  *Direction(self) = *pdir;
  *Haut(self) = haut;

  [self normalise];
}

- (void) updatePosition
{
  BOOL isCommandant = NO;
  float diffTeta;
  float x, y, rol;
  Vect3D wishedOmega = GGZeroVect;

  x = y = rol = 0;

  if( modeInertie & GGManche )
    {
      Vect3D 		v;
      real			beta;
      real			cosb, sinb;
      beta = speedRotation*realDeltaTime;
      if( beta < 1 )
	/* we do nothing otherwise because it turns too fast
	 */
	{
	  Vect3D		newDir;
	  mulScalVect(Haut(self), _cosam, &v);
	  addLambdaVect(&v, _sinam, Gauche(self), &v);

	  /*	  [theTableau addLogString: [NSString stringWithFormat:@"angle=%e",
		  angleManche]];

		  [theTableau addLogString: [NSString stringWithFormat:@"|v|=%e",
		  prodScal(&v, &v)]];
	  */
	  cosb = cos(beta);
	  sinb = sin(beta);
	  mulScalVect(Direction(self), cosb, &newDir);
	  addLambdaVect(&newDir, sinb, &v, &newDir);
	      
	  [self _adjustMatrixWithDirection: &newDir
		orthoDir: &v
		cosb: cosb
		sinb: sinb];
	}
    }
  if( modeInertie & GGWishDirection )
    {
      Vect3D 	tmp, v;
      real 		prodWishDir;
      real		beta;
      real		cosb, sinb;

      prodVect(&wishDir, Direction(self), &tmp);
      prodWishDir = prodScal(&tmp, &tmp);

      if( prodWishDir <= EPS )
	{
	  *Direction(self) = wishDir;
	  _flags.majRepere = YES;
	  [self stopWishDirection];
	}
      else
	{
	  beta = powerRotation*speedRotation*deltaTime;
	  prodVect(Direction(self), &tmp, &v);
	  normaliseVect(&v);
	  if( beta > 2 )
	    {
	      cosb = prodScal(&wishDir, Direction(self));
	      sinb = prodScal(&wishDir, &v);
	      [self _adjustMatrixWithDirection: &wishDir
		    orthoDir: &v
		    cosb: cosb
		    sinb: sinb];
	    }
	  else
	    {
	      real tmpcos;
	      Vect3D tdir;
	      cosb = cos(beta);
	      if( cosb < (tmpcos = prodScal(&wishDir, Direction(self))) )
		{
		  cosb = tmpcos;
		  sinb = prodScal(&wishDir, &v);
		  tdir = wishDir;
		}
	      else
		{
		  sinb = sin(beta);
		  mulScalVect(Direction(self), cosb, &tdir);
		  addLambdaVect(&tdir, sinb, &v, &tdir);
		}
	      [self _adjustMatrixWithDirection: &tdir
		    orthoDir: &v
		    cosb: cosb
		    sinb: sinb];
	    }
	}
    }
  diffTeta = realDeltaTime*speedRotation;
  if( modeInertie & GGLeft )
    x -= 1.0;
  if( modeInertie & GGRight )
    x += 1.0;
  if( modeInertie & GGUp )
    y += 1.0;
  if( modeInertie & GGDown )
    y -= 1.0;
  if( modeInertie & GGAccelere )
    addLambdaVect(Pousse(self), power*gAcc, Direction(self), Pousse(self));
  //	addLambdaVect(Speed(self), deltaTime*power*gAcc, Direction(self),
  //		      Speed(self));
  if( modeInertie & GGDecelere )
    addLambdaVect(Pousse(self), -power*gFrein, Direction(self), Pousse(self));
  //	addLambdaVect(Speed(self), -deltaTime*power*gFrein, Direction(self),
  //		      Speed(self));
  if( modeInertie & GGStrafLeft )
    rol += 1;
  if( modeInertie & GGStrafRight )
    rol -= 1;
  if( modeInertie & GGPousse )
    addLambdaVect(Pousse(self), power*gAcc, &pousse, Pousse(self));
  //	addLambdaVect(Speed(self), gAcc*power*deltaTime, 
  //		      &pousse, Speed(self));

  x += xMouvement;
  y -= yMouvement;
  rol -= rolMouvement;

  if( fabs(x) > 0.01 )
    {
      x = GGClamp(x);
      wishedOmega.y = -x;
      isCommandant = YES;
      //	  Momentum(self)->y -= x;
      //	  monAddLambdaVect(Direction(self), -x*diffTeta, Droite(self),
      //			   Direction(self));
    }
  if( fabs(y) > 0.01 )
    {
      y = GGClamp(y);
      wishedOmega.x = y;
      isCommandant = YES;
      //	  Momentum(self)->x += y;
      //	  monAddLambdaVect(Direction(self), -y*diffTeta, Haut(self),
      //			   Direction(self));
    }

  if( fabs(rol) > 0.01 )
    {
      rol = GGClamp(rol);
      wishedOmega.z = -rol;
      isCommandant = YES;
      //	  Momentum(self)->z += rol;
      //	  monAddLambdaVect(Haut(self), +diffTeta*rol, Gauche(self), 
      //			   Haut(self));	  
    }

  {
    Vect3D temp;
    double delta = (isCommandant ? realDeltaTime : deltaTime);

    mulScalVect(Pousse(self), deltaTime, &temp);
    addVect(Speed(self), &temp, Speed(self));

    if( thePref.DirectionHinter )
      {
	real normeDiff;
	diffVect(&wishedOmega, Omega(self), &temp);
	normeDiff = prodScal(&temp, &temp);
	if( normeDiff > EPS*EPS )
	  {
	    //		normeDiff = sqrt(normeDiff);
	    //		divScalVect(&temp, normeDiff, &temp);
	    dualProduct(&temp, &inertie, &temp);
	    normeDiff = prodScal(&temp, &temp);
	    if( normeDiff <= delta*delta )
	      divScalVect(&temp, delta, &temp);
	    else
	      divScalVect(&temp, sqrt(normeDiff), &temp);
	    addVect(Momentum(self), &temp, Momentum(self));
	  }
	else
	  *Omega(self) = wishedOmega;
      }
    else
      addVect(Momentum(self), &wishedOmega, Momentum(self));

    initVect(&temp, delta*(Momentum(self)->x)/inertie.x,
	     delta*(Momentum(self)->y)/inertie.y,
	     delta*(Momentum(self)->z)/inertie.z);

    addVect(Omega(self), &temp, Omega(self));

    SetZeroVect(*Pousse(self));
    SetZeroVect(*Momentum(self));
  }


  [super updatePosition];

}

//  - (void) genMove: (int) dx and: (int) dy
//  {
//    monAddLambdaVect(Direction(self), -0.005*dx, Droite(self), 
//  		Direction(self));
//    monAddLambdaVect(Direction(self), -0.005*dy, Haut(self), 
//  		Direction(self));
//  }

- (void) updateShipData
{
  NSDebugMLLogWindow(@"Carburant", @"%g %g", lastTime, gameTime);
  if( lastTime >= 0 )
    {
      if( carburant < 0 )
	{
	  [self notify: @"No more fuel"];
	  [self setPower: 0];
	  carburant = 0;
	}
      else if( carburant > 0)
	carburant -= power*consommation*(gameTime-lastTime);
    }
  lastTime = gameTime;
}

- (void) enterUniverse
{
  [theLoop slowSchedule:self
	   withSelector: @selector(updateShipData)
	   arg: nil];
  [super enterUniverse];
}

- (void) beginSchedule
{
  [super beginSchedule];
  if( _flags.locality == 1 )
    [theLoop slowSchedule: self
	     withSelector: @selector(tryLocalCoordinates:)
	     arg: nil];
  else if( _flags.locality == 2 )
    [theLoop slowSchedule: self
	     withSelector: @selector(tryGlobalCoordinates:)
	     arg: nil];
}

- (void) endSchedule
{
  [[theLoop fastScheduler]   deleteTarget: self
			    withSelector: @selector(updatePosition)];
  [[theLoop slowScheduler]
    deleteTarget: self
    withSelector: @selector(tryLocalCoordinates:)];
  [[theLoop slowScheduler]
    deleteTarget: self
    withSelector: @selector(tryGlobalCoordinates:)];
  [super endSchedule];
}

- (double) carburant
{
  return carburant;
}

- (void) setCarburant: (double) val
{
  carburant = val;
}

- (void) lockMissile
{
  if( nbrMissile == 0 )
    [self notify: @"No more Missile to lock"];
  else if( currentMissile >= 0 && missiles[currentMissile] != nil )
    {
      [self notify: @"missile %d unlock", currentMissile+1];
      currentMissile = -1;
    }
  else
    {
      int i;
      for(i = 0; i < MAXMISSILE; ++i)
	{
	  if( missiles[i] != nil )
	    {
	      currentMissile = i;
	      [self notify: @"missile %d lock", currentMissile+1];
	      break;
	    }
	}
      NSParameterAssert(i < MAXMISSILE);
    }
}

- (GGShip *) fireMissile
{
  GGShip 		*unShip;
  ShipErrorCode		code;

  if( currentMissile < 0 || missiles[currentMissile] == nil )
    {
      consoleLog(@"no missile lock");
      return nil;
    }

  code = [self removeItem: missiles[currentMissile]];

  NSAssert(code == NoError, NSInternalInconsistencyException);

  currentMissile = -1;

  unShip = [GGTetra spaceObjectAt: self];

  addLambdaVect(Point(unShip), -20, Haut(unShip), Point(unShip));

  addLambdaVect(Speed(unShip), 4, Direction(unShip), Speed(unShip));

  [theWorld addShip: unShip];
  [nodePere addSubNode: unShip];
  [unShip scheduleAsLocal];

  [unShip setIA: [AutoPilotCommon missilePilotWithTarget: target
				  andShip: unShip]];
  [unShip setName: @"Missile"];

  return unShip;
}

- (void) startFireLaser
{
  NSDebugMLLog(@"Laser", @"start firing laser %g", timeLaserRelaxation);
  if( laserGun[0] == nil )
    return;
  if( timeLaserRelaxation > 0 )
    {
      if( timeStartFire + timeLaserRelaxation < gameTime )
	{
	  [Laser fireLaserAt: self
		 power: 1e6*laserGun[0]->arg1
		 lifeTime: 0.2];
	  timeStartFire = gameTime;
	}
    }
  else
    {
      [laser destroy];
      ASSIGN(laser, [Laser fireLaserAt: self
			   power: 1e6*laserGun[0]->arg1
			   lifeTime: 0]);
    }
}

- (void) stopFireLaser
{
  NSDebugMLLog(@"Laser", @"stop firing laser");
  [laser destroy];
  DESTROY(laser);
}

- (float) energyForExplosion
{
  return masse;
}

- (void) doExplosion
{
  if( _flags.alive ) /*we are sure that a ship will explode only once !!!*/
    {
      NSDebugMLLog(@"Collision", @"%@ BOOOOOOMMM!", self);
      [self destroy];
      [theWorld genereExplosionFrom: self
		withEnergy: [self energyForExplosion]];
    }
}

- (void) chocWithEnergy: (real) val
{
  NSDebugMLLog(@"Collision", @"%@ Energie du choc : %g", self, val);
  if( invincible || endurance == 0)
    return;
  //  NSDebugMLLog(@"Collision", @"%@ dans la gueule", self);
  if( val < bouclier )
    bouclier -= val;
  else
    {
      val -= bouclier;
      bouclier = 0;
      if( val < endurance )
	endurance -= val;
      else
	{
	  endurance = 0;
	  [self doExplosion];
	}
    }
}


- (void) setMasseCoque: (float) theMasse
{
  masse = theMasse;
}

- (void) setInvincibility: (BOOL) flag
{
  invincible = flag;
}

- (void) scheduleAsLocal
{
  _flags.locality = 2;
  [self setVisible: YES];
  [[theLoop slowScheduler] deleteTarget: self
			  withSelector: @selector(tryLocalCoordinates:)];
  [theLoop slowSchedule: self
	   withSelector: @selector(tryGlobalCoordinates:)
	   arg: nil];
  [self _activate];
}

- (void) scheduleAsGlobal
{
  _flags.locality = 1;
  [self setVisible: NO];
  [[theLoop slowScheduler] deleteTarget: self
			  withSelector: @selector(tryGlobalCoordinates:)];
  [theLoop slowSchedule: self
	   withSelector: @selector(tryLocalCoordinates:)
	   arg: nil];
  [self _activate];
}

- (void) switchToLocal:(GGRepere *)_repere
{
  NSDebugMLLog(@"SpaceObj", @"switch to local");

  NSAssert((GGNode *)_repere != [self nodePere], @"already in local...");
  [_repere reparent: self];
  [self scheduleAsLocal];
}

- (void) adjustGlobalCoordinates
{
  GGSpaceObject *node = [manager repere];
  node = node->nodePere;
  while(node != nil && ! [node isInNeighboorhood: self]) node = node->nodePere;
  if( node != nil ) //so self isInNeighboorhood of node
    {
      if( node != nodePere )
	{
	  NSDebugMLLog(@"SpaceObj", @"change father orbiteur from %@ to %@",
		       nodePere, node);
	  [node reparent: self];
	}
    }
  NSDebugMLLogWindow(@"SpaceObj", @"node = %@ nodePere = %@", node, nodePere);
}

- (void) switchToGlobal: (GGRepere *) _repere
{
  NSDebugMLLog(@"SpaceObj", @"switch to global");

  NSAssert((GGNode *)_repere == [self nodePere], @"Object is not local");
  [self adjustGlobalCoordinates];
  //  [[_repere ancestor] reparent: self];
  [self scheduleAsGlobal];
}


- (void) tryLocalCoordinates: obj
{
  GGRepere *_repere = [manager repere];
  if( distanceAbsolu2(self, _repere) < distanceLocal2 )
    {
      [self switchToLocal: _repere];
    }
  else
    [self adjustGlobalCoordinates];
}

- (void) tryGlobalCoordinates: obj
{
  GGRepere * _repere = [manager repere];
  double d;
  d = (prodScal(Point(self), Point(self)));
  if( d > distanceGlobal2 )
    {
      [self switchToGlobal: _repere];
      //      NSLog(@"%@ go global", self);
    }
}

- (BOOL) shouldDrawMoreThings
{
  return YES;
}

- (void) rasteriseMoreThings
{
  GLUquadricObj *quad = gluNewQuadric();
  gluQuadricNormals(quad, GLU_NONE);
  glPushAttrib(GL_ENABLE_BIT | GL_DEPTH_BUFFER_BIT);

  glDisable(GL_LIGHTING);
  glDisable(GL_CULL_FACE);
  glBegin(GL_TRIANGLE_FAN);
  glColor3f(1, 1, 1);
  glVertex3f(posReacteur1.x, posReacteur1.y, posReacteur1.z);
  glColor3f(0.4+0.4*power, 0.4+0.4*power, 0.6+0.4*power);
  glVertex3f(posReacteur1.x-0.2, posReacteur1.y-0.2, posReacteur1.z);
  glVertex3f(posReacteur1.x-0.2, posReacteur1.y+0.2, posReacteur1.z);
  glVertex3f(posReacteur1.x+0.2, posReacteur1.y+0.2, posReacteur1.z);
  glVertex3f(posReacteur1.x+0.2, posReacteur1.y-0.2, posReacteur1.z);
  glVertex3f(posReacteur1.x-0.2, posReacteur1.y-0.2, posReacteur1.z);
  glEnd();

  glBegin(GL_TRIANGLE_FAN);
  glColor3f(1, 1, 1);
  glVertex3f(posReacteur2.x, posReacteur2.y, posReacteur2.z);
  glColor3f(0.4+0.4*power, 0.4+0.4*power, 0.6+0.4*power);
  glVertex3f(posReacteur2.x-0.2, posReacteur2.y-0.2, posReacteur2.z);
  glVertex3f(posReacteur2.x-0.2, posReacteur2.y+0.2, posReacteur2.z);
  glVertex3f(posReacteur2.x+0.2, posReacteur2.y+0.2, posReacteur2.z);
  glVertex3f(posReacteur2.x+0.2, posReacteur2.y-0.2, posReacteur2.z);
  glVertex3f(posReacteur2.x-0.2, posReacteur2.y-0.2, posReacteur2.z);
  glEnd();


  glEnable(GL_BLEND);
  glEnable(GL_CULL_FACE);
  glDepthMask(GL_FALSE);

  if( power >= 0.05 )
    {
      glBegin(GL_TRIANGLE_FAN);
      glColor4f(0.4, 0.4, 1,0);
      glVertex3f(posReacteur1.x, posReacteur1.y, posReacteur1.z-60*power);
      glColor4f(0.4+0.4*power, 0.4+0.4*power, 0.6+0.4*power, 0.8);
      glVertex3f(posReacteur1.x-0.2, posReacteur1.y-0.2, posReacteur1.z);
      glVertex3f(posReacteur1.x-0.2, posReacteur1.y+0.2, posReacteur1.z);
      glVertex3f(posReacteur1.x+0.2, posReacteur1.y+0.2, posReacteur1.z);
      glVertex3f(posReacteur1.x+0.2, posReacteur1.y-0.2, posReacteur1.z);
      glVertex3f(posReacteur1.x-0.2, posReacteur1.y-0.2, posReacteur1.z);
      glEnd();
  
      glBegin(GL_TRIANGLE_FAN);
      glColor4f(0.4, 0.4, 1,0);
      glVertex3f(posReacteur2.x, posReacteur2.y, posReacteur2.z-60*power);
      glColor4f(0.4+0.4*power, 0.4+0.4*power, 0.6+0.4*power, 0.8);
      glVertex3f(posReacteur2.x-0.2, posReacteur2.y-0.2, posReacteur2.z);
      glVertex3f(posReacteur2.x-0.2, posReacteur2.y+0.2, posReacteur2.z);
      glVertex3f(posReacteur2.x+0.2, posReacteur2.y+0.2, posReacteur2.z);
      glVertex3f(posReacteur2.x+0.2, posReacteur2.y-0.2, posReacteur2.z);
      glVertex3f(posReacteur2.x-0.2, posReacteur2.y-0.2, posReacteur2.z);
      glEnd();
  
    }

  glColor4f(0.7, 0.7, 1, 0.2*power);
  glPushMatrix();
  glTranslatef(posReacteur1.x, posReacteur1.y, posReacteur1.z);
  gluSphere(quad, 1, 10, 10);
  glPopMatrix();
  glPushMatrix();
  glTranslatef(posReacteur2.x, posReacteur2.y, posReacteur2.z);
  gluSphere(quad, 1, 10, 10);
  glPopMatrix();

  glPopAttrib();
  gluDeleteQuadric(quad);
}

- (void) rasterise
{
  [model rasterise];
}

- (void) addMomentum: (Vect3D *)v
	  resultante: (Vect3D *)w
{
  addVect(Momentum(self), v, Momentum(self));
  addVect(Pousse(self), w, Pousse(self));
}

- (void) addResultante: (Vect3D *)v
{
  addVect(Pousse(self), v, Pousse(self));
}

- (int) indexSystem
{
  return indexSystem;
}

- (void) setIndexSystem: (int) index
{
  indexSystem = index;
}

- (void) setTrack: (BOOL) val
{
  tracked = val;
  if( target != nil && [target isKindOfClass: [GGShip class]])
    [(GGShip *) target setTrack: val];
  [self _activate];
}

- (BOOL) tracked
{
  return tracked;
}
@end

@implementation GGShip (Private)
- (void) _activate
{
  if( _flags.locality == 2 || tracked )
    {
      [theLoop fastSchedule: self
	       withSelector: @selector(updatePosition)
	       arg: nil];
      if( ! _flags.activated )
	{
	  NSDebugMLLog(@"GGShip", @"%@ activated", self);
	  _flags.activated = 1;
	  [[NSNotificationCenter defaultCenter]
	    postNotificationName: GGNotificationShipBecomeActivated
	    object: self];
	}
    }
  else
    {
      [[theLoop fastScheduler]
	deleteTarget: self
	withSelector: @selector(updatePosition)];

      if( _flags.activated )
	{
	  NSDebugMLLog(@"GGShip", @"%@ inactivated", self);
	  _flags.activated = 0;
	  [[NSNotificationCenter defaultCenter]
	    postNotificationName: GGNotificationShipBecomeInactivated
	    object: self];
	}
    }
}

- (void) setModel: (ModelObject *) _model
{
  ASSIGN(model, _model);
  if( model != nil )
    {
      Vect3D *min, max;

      min = [model boundingBox: &max];
      initVect(&posReacteur1, -3.32, 0.44, min->z-1);
      initVect(&posReacteur2, 3.32, 0.44, min->z-1);
    }
}

@end

NSString *shipMsgFromErrorCode(ShipErrorCode code)
{
  switch(code)
    {
    case NoError: 
      return @"";
    case NoMoney:
      return @"Not enought Money";
    case GunBusy:
      return @"No Laser gun slot available";
    case MissilePylonBusy:
      return @"No rocket slot available";
    case   NoPlaceLeft:	
      return @"No space left";
    case  EquipmentNotFound:
      return @"Impossible, do a bug report!!!";
    default:
      return @"Impossible, do a bug report!!!";
    }
}
