/*
 *  SystemView.m
 *
 *  Copyright (c) 2001 by Frdric De Jaeger
 *  
 *  Date: June 2001
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <GL/gl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <Foundation/NSObject.h>
#include <Foundation/NSDebug.h>
#include <Foundation/NSException.h>
#include <Foundation/NSDictionary.h>
#include <Foundation/NSArray.h>

#include "utile.h"
#include "SystemView.h"
#include "GGNode.h"
#include "System.h"
#include "World.h"
#include "GGSpaceObject.h"
#include "GGOrbiteur.h"
#include "GGRepere.h"
#include "GGInput.h"
#include "Tableau.h"
#include "Preference.h"
#include "Console.h"
#include "GGWidget.h"

static const double foc = Focal;


@interface SystemView (Private)
- (void) centerRepereLocal;
- (void) updateProjection;
- (void) checkClosestOrbiteur;
- (void) _updatePositionCamera;
@end

@interface Orbite : GGSpaceObject
{
  float		angle;
  float		radius;
}
+ orbiteWithAngle: (float) _angle
	   radius: (float) _radius;
@end

@implementation SystemView
- (id) initWithFrame: (NSRect)frameRect
{
  System *currentSystem;
  [super initWithFrame: frameRect];
  [self enableTextureBackground];
  topNode = [GGNode new];
  [topNode setManager: self];
  cameraObj = [GGSpaceObject new];

  currentSystem = [theWorld currentSystem];

  mainStar = [[currentSystem mainStar] copy];

  leRepereLocal = [GGRepere new];
  //  [leRepereLocal beginSchedule];
  [leRepereLocal setDefaultName: @"repere local"];
  [topNode addSubNode: leRepereLocal];

  initVect(Point(cameraObj), 0, 0, 2e11);
  initVect(Direction(cameraObj), 0, 0, -1);
  initVect(Haut(cameraObj), 0, 1, 0);
  [cameraObj normalise];
  [topNode addSubNode: cameraObj];

  *Point(leRepereLocal) = *Point(cameraObj);
  [leRepereLocal reparent: cameraObj];


  console = [[Console alloc]
	      initWithFrame: NSMakeRect(0, 0, 0.3*frameRect.size.width,
					0.3*frameRect.size.height)];
  [console setTransluentBackground: YES];
  [self addSubview: console];

  [self centerRepereLocal];

  [self updateProjection];
  [self checkClosestOrbiteur];

  [theLoop fastSchedule: self
	   withSelector: @selector(updateView)
	   arg: nil];
  [theLoop slowSchedule: self
	   withSelector: @selector(checkClosestOrbiteur)
	   arg: nil];
  [theLoop slowSchedule: self
	   withSelector: @selector(centerRepereLocal)
	   arg: nil];

  NSParameterAssert(theClosest != nil);
  
  distance = sqrt(distanceAbsolu2(theClosest, cameraObj));

  [[self titleView]
    updateString: @"Current System"];

  return self;
}

- (void) updateProjection
{
  NSRect frameWindow = [[theGGInput contentView] frame];
  projectionParam.ratio = (double) frameWindow.size.width / 
    frameWindow.size.height;
  projectionParam.yFoc = tan((double)Focal/2.0/180*M_PI);
  projectionParam.xFoc = projectionParam.yFoc*projectionParam.ratio;
  projectionParam.zmin = 1;
  projectionParam.zmax = 10000;
}

- (void) open
{
  [super open];
  [_window makeFirstResponder: self];
}

- (void) dealloc
{
  [theLoop unSchedule: self];
  RELEASE(console);
  RELEASE(topNode);
  RELEASE(cameraObj);
  RELEASE(mainStar);
  RELEASE(leRepereLocal);
  NSDebugMLLog(@"SystemView", @"deallocation");
  [super dealloc];
}

- (void) centerRepereLocal
{
  if( prodScal(Point(cameraObj), Point(cameraObj)) > square(1e6) )
    {
      [leRepereLocal updateOrigine: cameraObj];
      //      NSDebugMLLog(@"SystemView", @"update Position");
    }
}

- (void) initGLBeforeDrawing
{
  Vect4D lightPosition = { {-1, 0, 1}, 0};
  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
  glLightfv(GL_LIGHT0, GL_POSITION, (real*)&lightPosition);

  /*ncessaire pour avoir une ilumination correcte des objets proches.
   */
  glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER,   (int) thePref.LightLocal);
  
  if( thePref.GLSmooth )
    glShadeModel(GL_SMOOTH);
  else
    glShadeModel(GL_FLAT);
}

- (void) myLockFocusInRect
{
  NSRect wrect;

  NSAssert(_window != nil, NSInternalInconsistencyException);

  wrect = [self convertRect: [self bounds] toView: nil];
  NSDebugLLog(@"NSView", @"Displaying rect \n\t%@\n\t window %p", 
	      NSStringFromRect(wrect), _window);
  [_window setGLFrameInRect: wrect];
}


- (void) drawRect: (NSRect) _theRect
{
  //  printf("%d %d %d %d\n", x, y, w, h);
  NSRect theRect = [self frame];
  ProjectionParam param;

  param.ratio = theRect.size.width/theRect.size.height;
  param.yFoc = tan((double)foc/2/180*M_PI);
  param.xFoc = param.yFoc*param.ratio;
  param.width = theRect.size.width;
  param.height = theRect.size.height;
  param.zmin = 1;
  param.zmax = 1e20;
  param.label = YES;
  projectionParam = param;

  [self _updatePositionCamera];

  glPushAttrib(GL_DEPTH_BUFFER_BIT | GL_VIEWPORT_BIT);
  glDepthFunc(GL_ALWAYS);
  glEnable(GL_DEPTH_TEST);
  glDepthMask(GL_TRUE);
  glDepthRange(1, 1);
  [super drawRect: _theRect];
  glPopAttrib();
  
  glEnable(GL_SCISSOR_TEST);
  glDisable(GL_DEPTH_TEST);
  glDepthMask(GL_TRUE);
  glDepthRange(0,1);
  
  
  [self myLockFocusInRect];

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(Focal,param.ratio,1,1e20);
  glMatrixMode(GL_MODELVIEW);

  [self initGLBeforeDrawing];
  [cameraObj drawGraphProjection: &param];
  glDisable(GL_LIGHTING);
  [_window initWindowGL];
}

+ systemView
{
  return AUTORELEASE([self new]);
}

- (void) makeFatherCurrentOrbiteur
{
  GGNode 	*newNode;
  GGNode 	*oldNode;
  GGOrbiteur	*orbiteurPere = [theClosest orbiteurPere];

  if( orbiteurPere == nil )
    return;

  [theClosest stopBeingOrigineOfLocale];
  oldNode = [theClosest nodePere];
  NSParameterAssert(oldNode != nil);
  newNode = [oldNode nodePere];
  NSParameterAssert(newNode != nil);
  
  [newNode reparent: leRepereLocal];
  [newNode reparent: theClosest];

  [[theClosest sons] makeObjectsPerformSelector: @selector(endSchedule)];

  [oldNode removeFromSuperNode];

  theClosest = [theClosest orbiteurPere];
  NSParameterAssert(theClosest != nil);

  [theClosest startBeingMainLocale];
}

- (void) _addOrbite: (NSArray *)sons
	     atNode: (GGNode *)newNode
{
  int n = [sons count];
  if( n > 0 )
    {
      GGOrbiteur *orbs[n];
      int i;

      [sons getObjects: orbs];
      for( i = 0; i < n; ++i)
	{
	  Orbite *o = [Orbite orbiteWithAngle: [orbs[i] anomalie]
			      radius: [orbs[i] parametresOrbitaux]->orbiteRadius];
	  NSDebugLLog(@"Orbite", @"radius %@ = %g", orbs[i],
		      [orbs[i] parametresOrbitaux]->orbiteRadius);
	  [newNode addSubNode: o];
	}
    }
}

- (void) makeSonCurrentOrbiteur: (GGOrbiteur *)son
{
  GGNode 	*newNode;
  GGNode 	*currentNode;
  NSArray 	*sons;

  [theClosest stopBeingMainLocale];
  currentNode = [theClosest nodePere];
  newNode = [GGNode new];
  [newNode setName: [NSString stringWithFormat: @"repere translation from %@",
			      son]];

  *Galileen(newNode) = *Galileen(son);
  
  [currentNode addSubNode: newNode];
  [newNode reparent: son];
  [newNode reparent: leRepereLocal];
  theClosest = son;

  [son startBeingOrigineOfLocale: newNode];
  sons = [son sons];
  [newNode addSubNodeFromArray: sons];
  [son beginSchedule];
  [sons makeObjectsPerformSelector: @selector(beginSchedule)];

  [self _addOrbite: sons
	atNode: newNode];

  RELEASE(newNode);
}

- (void) checkClosestOrbiteur
{
  GGOrbiteur *old = theClosest;
  if( theClosest == nil ) //first time
    {
      NSArray	*sons;
      GGOrbiteur *soleil = mainStar;
      theClosest = soleil;
      [topNode addSubNode: soleil];
      [soleil startBeingOrigineOfLocale: topNode];
      sons = [soleil sons];
      NSDebugLLog(@"SystemView", @"sons = %@", sons);
      [topNode addSubNodeFromArray: sons];
      [soleil beginSchedule];
      [sons makeObjectsPerformSelector: @selector(beginSchedule)];
      [self _addOrbite: sons
	    atNode: topNode];
    }
  else
    {
      Vect3D 		pos;
      GGOrbiteur	*son;
      //FIXME : source de bug pour le futur
      [cameraObj positionInRepere: [theClosest subSystemFrame]
		  in: &pos];
      NSDebugMLLogWindow(@"World", @"position orb %@", stringOfVect(&pos));
      if( prodScal(&pos, &pos) > square([theClosest tailleSysteme]) )
	[self makeFatherCurrentOrbiteur];
      else if ((son = [theClosest closestFromRelative: &pos] ) != theClosest )
	[self makeSonCurrentOrbiteur: son];
      /* else do nothing */
    }
  if( currentOrbiteur == nil )
    currentOrbiteur = theClosest;
  if( theClosest != old && theClosest != nil )
    {
      [console clear];
      [console log: @"relative to %@", theClosest];
    }
}

- (GGSpaceObject *)camera
{
  return cameraObj;
}

- (void) centerView
{
  NSLog(@"Center View");
}

- (void) keyDown: (Event *)pev
{
  switch(pev->key.keysym.sym)
    {
    case SDLK_c:
      [self centerView];
      break;
    case SDLK_PAGEUP:
      state |= GGAccelere;
      break;
    case SDLK_PAGEDOWN:
      state |= GGDecelere;
      break;
    default:
      [super keyDown: pev];
      break;
    }
}

- (void) keyUp: (Event *)pev
{
  switch(pev->key.keysym.sym)
    {
    case SDLK_PAGEUP:
      state &= ~GGAccelere;
      break;
    case SDLK_PAGEDOWN:
      state &= ~GGDecelere;
      break;
    default:
      [super keyDown: pev];
      break;
    }
}


- (void) mouseDown: (Event *) pev
{
  if( pev->button.button == SDL_BUTTON_RIGHT )
    _drag = YES;
  else if( pev->button.button == SDL_BUTTON_LEFT )
    {
      NSPoint 		res 	= 
	[self convertPoint: NSMakePoint(pev->button.x, pev->button.y)
	      fromView: nil];
      GGOrbiteur 	*orbCliq = nil;
      int 		n 	= [[theClosest sons] count];
      GGOrbiteur 	*lstobj[n+1];
      int 		distCliq = 10000;
      Camera 		c;
      GGGalileen 	gal;
      int 		i;
      
      [theClosest galileanTransformationTo: cameraObj
		  in: &gal];

      c.modelView = gal.matrice;
      c.param = &projectionParam;

      lstobj[0] = theClosest;
      [[theClosest sons] getObjects: lstobj+1];

      NSDebugMLLog(@"SystemView", @"mouse %g %g", res.x, res.y);


      for(i = 0; i < n+1; ++i)
	{
	  Vect3D screenPos;
	  //	  if( getWinCoordinates(&c, Point(lstobj[i]), &screenPos) )
	  if( [lstobj[i] windowsCoordinatesForCamera: cameraObj
		     parameter: &projectionParam
		     result: &screenPos] )
		     
	    {
	      int temp = fabs(screenPos.x-res.x) + fabs(screenPos.y-res.y);
	      NSDebugMLLog(@"SystemView", @"pos %@ %g %g", lstobj[i],
			   screenPos.x, screenPos.y);
	      if( temp < 20 && temp < distCliq )
		{
		  distCliq = temp;
		  orbCliq = lstobj[i];
		}
	    }
	}
      if ( orbCliq != nil )
	{
	  currentOrbiteur = orbCliq;
	}
    }
}

- (void) mouseUp: (Event *)pev
{
  if( pev->button.button == SDL_BUTTON_RIGHT )
    _drag = NO;
}

- (void) _updatePositionCamera
{
  real taille;
  PetitGalileen pg;
  taille = 2*Dimension(currentOrbiteur);
  if( distance < taille )
    distance = taille;
  [currentOrbiteur positionSpeedRelativesTo: cameraObj
		   decal: NULL
		   in: &pg];

  addVect(Point(cameraObj), &pg.point, Point(cameraObj));
  addLambdaVect(Point(cameraObj), -distance, Direction(cameraObj),
		Point(cameraObj));
}
- (void) mouseMoved: (Event *) pev
{
  if( _drag )
    {
      int dx = pev->motion.xrel;
      int dy = pev->motion.yrel;
      addLambdaVect(Haut(cameraObj), -0.005*dx, Gauche(cameraObj),
		    Haut(cameraObj));
      addLambdaVect(Direction(cameraObj), 0.005*dy, Haut(cameraObj),
		    Direction(cameraObj));
      [cameraObj normalise];
    }
}

- (void) updateView
{
  if( state )
    {
      if( state & GGAccelere && distance > 500 )
	distance *= pow(10,realDeltaTime/2);
      if( state & GGDecelere )
	distance *= pow(10,-realDeltaTime/2);

    }
}

- (GGRepere *)repere
{
  return leRepereLocal;
}

@end

@implementation Orbite
- initAngle: (float) _angle
     radius: (float) _radius
{
  [self init];
  angle = _angle;
  radius = _radius;
  _flags.graphical = YES;
  [self setVisible: YES];
  return self;
}

- (BOOL) shouldBeClipped
{
  return NO;
}

- (BOOL) shouldCache
{
  return YES;
}

+ orbiteWithAngle: (float) _angle
	   radius: (float) _radius
{
  Orbite *or = [self alloc];
  return AUTORELEASE([or initAngle: _angle
			 radius: _radius]);
}

#if 0
- (void) dealloc
{
  NSWarnMLog(@"on dgage ces orbites");
  [super dealloc];
}
#endif

- (void) rasterise: (Camera *)cam
{
  int i;
  //  NSLog(@"haa");
  glColor3f(0, 1, 0);
  glDisable(GL_LIGHTING);
  glBegin(GL_LINE_STRIP);
  for( i = 1; i < NBRSEGMENTINCIRCLE; ++i )
    {
      float a = angle + i*2*M_PI/NBRSEGMENTINCIRCLE;
      glVertex2f(radius*cos(a), radius*sin(a));
    }
  glEnd();
  glEnable(GL_LIGHTING);
}

@end
